"""FIFE modelers based on LightGBM, which trains gradient-boosted trees."""

import json
from typing import List, Union
from warnings import warn

import dask
from fife.base_modelers import (
    default_subset_to_all,
    Modeler,
    SurvivalModeler,
    StateModeler,
    ExitModeler,
)
import lightgbm as lgb
import numpy as np
import optuna
import pandas as pd
import shap


class LGBModeler(Modeler):
    """Train a gradient-boosted tree model for each lead length using LightGBM.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast.
        model (list): A trained LightGBM model (lgb.basic.Booster) for each
            lead length.
        objective (str): The LightGBM model objective appropriate for the
            outcome type, which is "binary" for binary classification.
        num_class (int): The num_class LightGBM parameter, which is 1 for
            binary classification.
    """

    def hyperoptimize(
        self,
        n_trials: int = 64,
        rolling_validation: bool = True,
        subset: Union[None, pd.core.series.Series] = None,
    ) -> dict:
        """Search for hyperparameters with greater out-of-sample performance.

        Args:
            n_trials: The number of hyperparameter sets to evaluate for each
                time horizon. Return None if non-positive.
            rolling_validation: Whether or not to evaluate performance on the
                most recent possible periods instead of the validation set
                labeled by self.validation_col. Ignored for a given time horizon
                if there is only one possible period for training and evaluation.
            subset:  A Boolean Series that is True for observations on which
                to train and validate. If None, default to all observations not
                flagged by self.test_col or self.predict_col.

        Returns:
            A dictionary containing the best-performing parameter dictionary for
            each time horizon.
        """

        def evaluate_params(
            trial: optuna.trial.Trial,
            train_data: lgb.Dataset,
            validation_data: lgb.Dataset,
        ) -> Union[None, dict]:
            """Compute out-of-sample performance for a parameter set."""
            params = {}
            params["num_iterations"] = trial.suggest_int("num_iterations", 8, 128)
            params["learning_rate"] = trial.suggest_uniform(
                "learning_rate", 2 ** -5, 0.5
            )
            params["num_leaves"] = trial.suggest_int("num_leaves", 8, 256)
            params["max_depth"] = trial.suggest_int("max_depth", 4, 32)
            params["min_data_in_leaf"] = trial.suggest_int("min_data_in_leaf", 4, 512)
            params["min_sum_hessian_in_leaf"] = trial.suggest_uniform(
                "min_sum_hessian_in_leaf", 2 ** -5, 0.25
            )
            params["bagging_freq"] = trial.suggest_int("bagging_freq", 0, 1)
            params["bagging_fraction"] = trial.suggest_uniform(
                "bagging_fraction", 0.5, 1
            )
            params["feature_fraction"] = trial.suggest_uniform(
                "feature_fraction", 0.5, 1
            )
            params["lambda_l1"] = trial.suggest_uniform("lambda_l1", 0, 64)
            params["lambda_l2"] = trial.suggest_uniform("lambda_l2", 0, 64)
            params["min_gain_to_split"] = trial.suggest_uniform(
                "min_gain_to_split", 0, 0.25
            )
            params["min_data_per_group"] = trial.suggest_int(
                "min_data_per_group", 1, 512
            )
            params["max_cat_threshold"] = trial.suggest_int("max_cat_threshold", 1, 512)
            params["cat_l2"] = trial.suggest_uniform("cat_l2", 0, 64)
            params["cat_smooth"] = trial.suggest_uniform("cat_smooth", 0, 2048)
            params["max_cat_to_onehot"] = trial.suggest_int("max_cat_to_onehot", 1, 64)
            params["max_bin"] = trial.suggest_int("max_bin", 32, 1024)
            params["min_data_in_bin"] = trial.suggest_int("min_data_in_bin", 1, 64)
            params["objective"] = self.objective
            params["num_class"] = self.num_class
            params["verbosity"] = -1
            booster = lgb.Booster(params=params, train_set=train_data)
            booster.add_valid(validation_data, "validation_set")
            for step in range(params["num_iterations"]):
                booster.update()
                validation_loss = booster.eval_valid()[0][2]
                trial.report(validation_loss, step)
                if trial.should_prune():
                    raise optuna.exceptions.TrialPruned()
            return validation_loss

        default_params = {
            "objective": self.objective,
            "num_iterations": 100,
            "num_class": self.num_class,
        }
        if n_trials <= 0:
            return {
                time_horizon: default_params for time_horizon in range(self.n_intervals)
            }
        params = {}
        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]
        for time_horizon in range(self.n_intervals):
            data = self.label_data(time_horizon)
            data = data[subset]
            data = self.subset_for_training_horizon(data, time_horizon)
            if rolling_validation:
                rolling_validation_n_periods = int(
                    max(
                        self.config.get("VALIDATION_SHARE", 0.25)
                        * data[self.config["TIME_IDENTIFIER"]].nunique(),
                        1,
                    )
                )
                rolling_validation_period_cutoff = pd.factorize(
                    data[self.config["TIME_IDENTIFIER"]], sort=True
                )[1][-rolling_validation_n_periods]
                rolling_validation_subset = (
                    data[self.config["TIME_IDENTIFIER"]]
                    >= rolling_validation_period_cutoff
                )
            else:
                rolling_validation_subset = [False] * data.shape[0]
            if (data[rolling_validation_subset]["_label"].nunique() > 1) and (
                data[~rolling_validation_subset]["_label"].nunique() > 1
            ):
                validation_subset = rolling_validation_subset
            else:
                validation_subset = data[self.validation_col]
            train_data = lgb.Dataset(
                data[~validation_subset][
                    self.categorical_features + self.numeric_features
                ],
                label=data[~validation_subset]["_label"],
                weight=data[~validation_subset][self.weight_col]
                if self.weight_col
                else None,
            )
            validation_data = train_data.create_valid(
                data[validation_subset][
                    self.categorical_features + self.numeric_features
                ],
                label=data[validation_subset]["_label"],
                weight=data[validation_subset][self.weight_col]
                if self.weight_col
                else None,
            )
            study = optuna.create_study(
                pruner=optuna.pruners.MedianPruner(),
                sampler=optuna.samplers.TPESampler(seed=self.config.get("SEED", 9999)),
            )
            study.optimize(
                lambda trial: evaluate_params(trial, train_data, validation_data),
                n_trials=n_trials,
            )
            params[time_horizon] = study.best_params
            params[time_horizon]["objective"] = self.objective
            default_booster = lgb.Booster(params=default_params, train_set=train_data)
            default_booster.add_valid(validation_data, "validation_set")
            for _ in range(default_params["num_iterations"]):
                default_booster.update()
            default_validation_loss = default_booster.eval_valid()[0][2]
            if default_validation_loss <= study.best_value:
                params[time_horizon] = default_params
        return params

    def build_model(
        self,
        n_intervals: Union[None, int] = None,
        params: dict = None,
        parallelize: bool = True,
    ) -> None:
        """Train and store a sequence of gradient-boosted tree models."""
        if n_intervals:
            self.n_intervals = n_intervals
        else:
            self.n_intervals = self.set_n_intervals()
        early_stopping = (params is None) or (
            any(["num_iterations" not in d for d in params.values()])
        )
        self.model = self.train(
            params=params,
            validation_early_stopping=early_stopping,
            parallelize=parallelize,
        )

    def train(
        self,
        params: Union[None, dict] = None,
        subset: Union[None, pd.core.series.Series] = None,
        validation_early_stopping: bool = True,
        parallelize: bool = True,
    ) -> List[lgb.basic.Booster]:
        """Train a LightGBM model for each lead length."""
        models = []
        if parallelize:
            for time_horizon in range(self.n_intervals):
                model = dask.delayed(self.train_single_model)(
                    time_horizon=time_horizon,
                    params=params,
                    subset=subset,
                    validation_early_stopping=validation_early_stopping,
                )
                models.append(model)
            models = dask.compute(*models)
        else:
            for time_horizon in range(self.n_intervals):
                model = self.train_single_model(
                    time_horizon=time_horizon,
                    params=params,
                    subset=subset,
                    validation_early_stopping=validation_early_stopping,
                )
                models.append(model)
        return models

    def train_single_model(
        self,
        time_horizon: int,
        params: Union[None, dict] = None,
        subset: Union[None, pd.core.series.Series] = None,
        validation_early_stopping: bool = True,
    ) -> lgb.basic.Booster:
        """Train a LightGBM model for a single lead length."""
        if params is None:
            params = {
                time_horizon: {
                    "objective": self.objective,
                    "num_iterations": self.config.get("MAX_EPOCHS", 256),
                }
            }
        params[time_horizon]["num_class"] = self.num_class
        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]
        data = self.label_data(time_horizon)
        data = data[subset]
        data = self.subset_for_training_horizon(data, time_horizon)
        if validation_early_stopping:
            train_data = lgb.Dataset(
                data[~data[self.validation_col]][
                    self.categorical_features + self.numeric_features
                ],
                label=data[~data[self.validation_col]]["_label"],
                weight=data[~data[self.validation_col]][self.weight_col]
                if self.weight_col
                else None,
            )
            validation_data = train_data.create_valid(
                data[data[self.validation_col]][
                    self.categorical_features + self.numeric_features
                ],
                label=data[data[self.validation_col]]["_label"],
                weight=data[data[self.validation_col]][self.weight_col]
                if self.weight_col
                else None,
            )
            model = lgb.train(
                params[time_horizon],
                train_data,
                early_stopping_rounds=self.config.get("PATIENCE", 4),
                valid_sets=[validation_data],
                valid_names=["validation_set"],
                categorical_feature=self.categorical_features,
                verbose_eval=True,
            )
        else:
            data = lgb.Dataset(
                data[self.categorical_features + self.numeric_features],
                label=data["_label"],
                weight=data[self.weight_col] if self.weight_col else None,
            )
            model = lgb.train(
                params[time_horizon],
                data,
                categorical_feature=self.categorical_features,
                verbose_eval=True,
            )
        return model

    def predict(
        self, subset: Union[None, pd.core.series.Series] = None, cumulative: bool = True
    ) -> np.ndarray:
        """Use trained LightGBM models to predict the outcome for each observation and time horizon.

        Args:
            subset: A Boolean Series that is True for observations for which
                predictions will be produced. If None, default to all
                observations.
            cumulative: If True, produce cumulative survival probabilies.
                If False, produce marginal survival probabilities (i.e., one
                minus the hazard rate).

        Returns:
            A numpy array of predictions by observation and lead
            length.
        """
        subset = default_subset_to_all(subset, self.data)
        predict_data = self.data[self.categorical_features + self.numeric_features][
            subset
        ]
        predictions = np.array(
            [
                lead_specific_model.predict(predict_data)
                for lead_specific_model in self.model
            ]
        ).T
        if cumulative:
            predictions = np.cumprod(predictions, axis=1)
        return predictions

    def transform_features(self) -> pd.DataFrame:
        """Transform features to suit model training."""
        data = self.data.copy(deep=True)
        if self.config.get("DATETIME_AS_DATE", True):
            date_cols = data.select_dtypes("datetime").columns + [
                col
                for col in data.select_dtypes("category")
                if np.issubdtype(data[col].cat.categories.dtype, np.datetime64)
            ]
            for col in date_cols:
                data[col] = (
                    data[col].dt.year * 10000
                    + data[col].dt.month * 100
                    + data[col].dt.day
                )
        else:
            data[data.select_dtypes("datetime")] = pd.to_numeric(
                data[data.select_dtypes("datetime")]
            )
            for col in data.select_dtypes("category"):
                if np.issubdtype(data[col].cat.categories.dtype, np.datetime64):
                    data[col] = data[col].astype(int)
        return data

    def save_model(self, file_name: str = "GBT_Model", path: str = "") -> None:
        """Save the horizon-specific LightGBM models that comprise the model to disk."""
        for i, lead_specific_model in enumerate(self.model):
            with open(f"{path}{i + 1}-lead_{file_name}.json", "w") as file:
                json.dump(lead_specific_model.dump_model(), file, indent=4)

    def compute_shap_values(
        self, subset: Union[None, pd.core.series.Series] = None
    ) -> dict:
        """Compute SHAP values by lead length, observation, and feature."""
        subset = default_subset_to_all(subset, self.data)
        shap_subset = self.data[self.categorical_features + self.numeric_features][
            subset
        ]
        shap_values = {
            str(i + 1)
            + "_lead": shap.TreeExplainer(lead_specific_model).shap_values(shap_subset)[
                0
            ]
            for i, lead_specific_model in enumerate(self.model)
            if lead_specific_model.num_trees() > 0
        }
        return shap_values


class LGBSurvivalModeler(LGBModeler, SurvivalModeler):
    """Use LightGBM to forecast probabilities of being observed in future periods."""

    pass


class GradientBoostedTreesModeler(LGBSurvivalModeler):
    """Deprecated alias for LGBSurvivalModeler"""

    warn(
        "The name 'GradientBoostedTreesModeler' is deprecated. "
        "Please use 'LGBSurvivalModeler' instead.",
        DeprecationWarning,
    )


class LGBStateModeler(LGBModeler, StateModeler):
    """Use LightGBM to forecast the future value of a feature conditional on survival."""

    pass


class LGBExitModeler(LGBModeler, ExitModeler):
    """Use LightGBM to forecast the circumstance of exit conditional on exit."""

    pass



"""FIFE modelers based on TensorFlow, which trains neural networks."""

from inspect import getfullargspec
from typing import List, Union
from warnings import warn

from fife.base_modelers import default_subset_to_all, Modeler, SurvivalModeler
from fife.nnet_survival import make_surv_array, surv_likelihood, PropHazards
import numpy as np
import optuna
import pandas as pd
import shap
import tensorflow as tf
import tensorflow.keras as keras
import tensorflow.keras.backend as K
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.layers import (
    Concatenate,
    Dense,
    Dropout,
    Embedding,
    Flatten,
    Input,
    Lambda,
    Layer,
)
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.regularizers import l2


def binary_encode_feature(col: pd.core.series.Series) -> pd.core.frame.DataFrame:
    """Map whole numbers to bits.

    Args:
        col: a pandas Series of whole numbers.

    Returns:
        A pandas DataFrame of Boolean values, each combination of values unique
        to each unique value in the given Series.
    """
    bits = int(np.log2(max(np.max(col), 1)) + 1)
    binary_columns = pd.DataFrame()
    for i in reversed(range(bits)):
        digit_value = 2 ** i
        binary_columns[f"{col.name}_bit{i}"] = col >= digit_value
        col = col % digit_value
    return binary_columns


def split_categorical_features(
    data: pd.core.frame.DataFrame,
    categorical_features: List[str],
    numeric_features: List[str],
) -> List[Union[pd.core.series.Series, pd.core.frame.DataFrame]]:
    """Split each categorical column in a DataFrame into its own list item.

    Necessary to specify inputs to a neural network with embeddings.

    Args:
        df: A DataFrame.
        categorical_features: A list of column names to split out.
        numeric_features: A list of column names to keep in a single DataFrame.

    Returns:
        A list where each element but the last is a Series and the last element
        is a DataFrame of the given numeric features.
    """
    return [data[col].cat.codes for col in categorical_features] + [
        data[numeric_features]
    ]


def freeze_embedding_layers(model: keras.Model) -> keras.Model:
    """Prevent embedding layers of the given neural network from training."""
    new_model = Model(model.inputs, model.outputs)
    for layer in new_model.layers:
        if type(layer).__name__ == "Embedding":
            layer.trainable = False
    return new_model


def make_predictions_cumulative(model: keras.Model) -> keras.Model:
    """Append a cumulative product layer to a neural network."""
    if isinstance(model.layers[-1], CumulativeProduct):
        return model
    cumprod_layer = CumulativeProduct()(model.output)
    return Model(model.inputs, cumprod_layer)


def make_predictions_marginal(model: keras.Model) -> keras.Model:
    """Remove final layer of a neural network if a cumulative product layer."""
    if isinstance(model.layers[-1], CumulativeProduct):
        return Model(model.inputs, model.layers[-2].output)
    return model


class CumulativeProduct(Layer):
    """Transform an array into its row-wise cumulative products."""

    def call(self, inputs, **kwargs):
        """Multiply each value with all previous values in the same row."""
        return K.cumprod(inputs, axis=1)


class TFModeler(Modeler):
    """Train a neural network model using Keras with TensorFlow backend.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of one-period intervals any
            individual is observed to survive.
        model (keras.Model): A trained neural network.
    """

    def hyperoptimize(
        self,
        n_trials: int = 64,
        subset: Union[None, pd.core.series.Series] = None,
        max_epochs: int = 128,
    ) -> dict:
        """Search for hyperparameters with greater out-of-sample performance.

        Args:
            n_trials: The number of hyperparameter sets to evaluate for each
                time horizon. Return None if non-positive.
            subset:  A Boolean Series that is True for observations on which
                to train and validate. If None, default to all observations not
                flagged by self.test_col or self.predict_col.

        Returns:
            A dictionary containing the best-performing parameters.
        """

        def evaluate_params(
            trial: optuna.trial.Trial,
            x_train: List[np.array],
            y_train: np.array,
            x_valid: List[np.array],
            y_valid: np.array,
            train_weights: np.array,
            valid_weights: np.array,
            max_epochs: int,
        ) -> Union[None, dict]:
            """Compute out-of-sample performance for a parameter set."""
            params = {}
            params["BATCH_SIZE"] = trial.suggest_int(
                "BATCH_SIZE", min(32, x_train[0].shape[0]), x_train[0].shape[0]
            )
            params["DENSE_LAYERS"] = trial.suggest_int("DENSE_LAYERS", 0, 3)
            params["DROPOUT_SHARE"] = trial.suggest_uniform("DROPOUT_SHARE", 0, 0.5)
            params["EMBED_EXPONENT"] = trial.suggest_uniform("EMBED_EXPONENT", 0, 0.25)
            params["EMBED_L2_REG"] = trial.suggest_uniform("EMBED_L2_REG", 0, 16.0)
            if self.categorical_features:
                params["POST_FREEZE_EPOCHS"] = trial.suggest_int(
                    "POST_FREEZE_EPOCHS", 4, max_epochs
                )
            else:
                params["POST_FREEZE_EPOCHS"] = 0
            max_pre_freeze_epochs = max_epochs
            params["PRE_FREEZE_EPOCHS"] = trial.suggest_int(
                "PRE_FREEZE_EPOCHS", 4, max_epochs
            )
            params["NODES_PER_DENSE_LAYER"] = trial.suggest_int(
                "NODES_PER_DENSE_LAYER", 16, 1024
            )
            construction_args = getfullargspec(self.construct_embedding_network).args
            self.model = self.construct_embedding_network(
                **{k.lower(): v for k, v in params.items() if k in construction_args}
            )
            self.data[self.numeric_features] = self.data[self.numeric_features].fillna(
                self.config.get("NON_CAT_MISSING_VALUE", -1)
            )
            model = self.construct_embedding_network(
                **{
                    k.lower(): v
                    for k, v in self.config.items()
                    if k in construction_args
                }
            )
            model.compile(
                loss=surv_likelihood(self.n_intervals), optimizer=Adam(amsgrad=True)
            )
            for step in range(params["PRE_FREEZE_EPOCHS"]):
                model.fit(
                    x_train,
                    y_train,
                    batch_size=params["BATCH_SIZE"],
                    epochs=1,
                    sample_weight=train_weights,
                )
                validation_loss = model.evaluate(
                    x_valid, y_valid, sample_weight=valid_weights
                )
                trial.report(validation_loss, step)
                if trial.should_prune():
                    raise optuna.exceptions.TrialPruned()
            model = freeze_embedding_layers(model)
            model.compile(
                loss=surv_likelihood(self.n_intervals), optimizer=Adam(amsgrad=True)
            )
            for step in range(params["POST_FREEZE_EPOCHS"]):
                model.fit(
                    x_train,
                    y_train,
                    batch_size=params["BATCH_SIZE"],
                    epochs=1,
                    sample_weight=train_weights,
                )
                validation_loss = model.evaluate(
                    x_valid, y_valid, sample_weight=valid_weights
                )
                trial.report(validation_loss, step + max_pre_freeze_epochs)
                if trial.should_prune():
                    raise optuna.exceptions.TrialPruned()
            return validation_loss

        default_params = {"BATCH_SIZE": 512, "PRE_FREEZE_EPOCHS": 16}
        if n_trials <= 0:
            return {
                time_horizon: default_params for time_horizon in range(self.n_intervals)
            }
        params = {}
        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]
        train_subset = subset & ~self.data[self.validation_col]
        valid_subset = subset & self.data[self.validation_col]
        x_train = self.format_input_data(subset=train_subset)
        x_valid = self.format_input_data(subset=valid_subset)
        y_surv = make_surv_array(
            self.data[[self.duration_col, self.max_lead_col]].min(axis=1),
            self.data[self.event_col],
            np.arange(self.n_intervals + 1),
        )
        y_train = y_surv[train_subset]
        y_valid = y_surv[valid_subset]
        train_weights = data[train_subset][self.weight_col] if self.weight_col else None
        valid_weights = data[valid_subset][self.weight_col] if self.weight_col else None
        study = optuna.create_study(
            pruner=optuna.pruners.MedianPruner(),
            sampler=optuna.samplers.TPESampler(seed=self.config.get("SEED", 9999)),
        )
        study.optimize(
            lambda trial: evaluate_params(
                trial,
                x_train,
                y_train,
                x_valid,
                y_valid,
                train_weights,
                valid_weights,
                max_epochs,
            ),
            n_trials=n_trials,
        )
        params = study.best_params
        if self.categorical_features:
            default_params["POST_FREEZE_EPOCHS"] = 16
        else:
            default_params["POST_FREEZE_EPOCHS"] = 0
        construction_args = getfullargspec(self.construct_embedding_network).args
        default_model = self.construct_embedding_network(
            **{
                k.lower(): v
                for k, v in default_params.items()
                if k in construction_args
            }
        )
        default_model.compile(
            loss=surv_likelihood(self.n_intervals), optimizer=Adam(amsgrad=True)
        )
        default_model.fit(
            x_train,
            y_train,
            batch_size=default_params["BATCH_SIZE"],
            epochs=default_params["PRE_FREEZE_EPOCHS"],
            sample_weight=train_weights,
        )
        default_model = freeze_embedding_layers(default_model)
        default_model.compile(
            loss=surv_likelihood(self.n_intervals), optimizer=Adam(amsgrad=True)
        )
        default_model.fit(
            x_train,
            y_train,
            batch_size=default_params["BATCH_SIZE"],
            epochs=default_params["POST_FREEZE_EPOCHS"],
            sample_weight=train_weights,
        )
        default_validation_loss = default_model.evaluate(
            x_valid, y_valid, sample_weight=valid_weights
        )
        if default_validation_loss <= study.best_value:
            params = default_params
        return params

    def build_model(
        self, n_intervals: Union[None, int] = None, params: dict = None
    ) -> None:
        """Train and store a neural network, freezing embeddings midway."""
        tf.random.set_seed(self.config.get("SEED", 9999))
        if params is None:
            params = self.config
        if n_intervals:
            self.n_intervals = n_intervals
        else:
            self.n_intervals = self.set_n_intervals()
        self.data[self.numeric_features] = self.data[self.numeric_features].fillna(
            self.config.get("NON_CAT_MISSING_VALUE", -1)
        )
        construction_args = getfullargspec(self.construct_embedding_network).args
        self.model = self.construct_embedding_network(
            **{k.lower(): v for k, v in params.items() if k in construction_args}
        )
        pre_freeze_early_stopping = "PRE_FREEZE_EPOCHS" not in params.keys()
        post_freeze_early_stopping = "POST_FREEZE_EPOCHS" not in params.keys()
        if post_freeze_early_stopping:
            params["POST_FREEZE_EPOCHS"] = params.get("MAX_EPOCHS", 256)
        if not pre_freeze_early_stopping:
            params["MAX_EPOCHS"] = params["PRE_FREEZE_EPOCHS"]
        self.model = self.train(
            params, validation_early_stopping=pre_freeze_early_stopping
        )
        if self.categorical_features:
            self.model = freeze_embedding_layers(self.model)
            params["MAX_EPOCHS"] = params["POST_FREEZE_EPOCHS"]
            self.model = self.train(
                params, validation_early_stopping=post_freeze_early_stopping
            )
        self.model = make_predictions_cumulative(self.model)

    def construct_embedding_network(
        self,
        dense_layers: int = 2,
        nodes_per_dense_layer: int = 512,
        dropout_share: float = 0.25,
        embed_exponent: float = 0,
        embed_L2_reg: float = 2.0,
    ) -> keras.Model:
        """Set embedding layers followed by alternating dropout/dense layers.

        Each categorical feature passes through its own embedding layer, which
        maps whole numbers to the real line.

        Each dense layer has a sigmoid activation function. The output layer
        has one node for each lead length.

        Args:
            dense_layers: The number of dense layers in the neural network.
            nodes_per_dense_layer: The number of nodes per dense layer in the neural network.
            dropout_share: The probability of a densely connected node of the neural network being set to zero weight during training.
            embed_exponent: The ratio of the natural logarithm of the number of embedded values to the natural logarithm of the number of unique categories for each categorical feature.
            embed_L2_reg: The L2 regularization coefficient for each embedding layer.

        Returns:
            An untrained Keras model.
        """
        if self.categorical_features:
            categorical_input_layers = [
                Input(shape=(1,), dtype="int32") for _ in self.categorical_features
            ]
            embed_layers = [
                Embedding(
                    input_dim=self.data[col].nunique() + 2,
                    output_dim=int((self.data[col].nunique() + 2) ** embed_exponent),
                    embeddings_regularizer=l2(embed_L2_reg),
                )(lyr)
                for (col, lyr) in zip(
                    self.categorical_features, categorical_input_layers
                )
            ]
            flatten_layers = [Flatten()(lyr) for lyr in embed_layers]
            numeric_input_layer = Input(shape=(len(self.numeric_features),))
            concat_layer = Concatenate(axis=1)(flatten_layers + [numeric_input_layer])
            dropout_layer = Dropout(dropout_share)(concat_layer)
        else:
            categorical_input_layers = []
            numeric_input_layer = Input(shape=(len(self.numeric_features),))
            dropout_layer = Dropout(dropout_share)(numeric_input_layer)
        for _ in range(dense_layers):
            dense_layer = Dense(nodes_per_dense_layer, activation="sigmoid")(
                dropout_layer
            )
            dropout_layer = Dropout(dropout_share)(dense_layer)
        output_layer = Dense(self.n_intervals, activation="sigmoid", name="output")(
            dropout_layer
        )
        model = Model(
            inputs=categorical_input_layers + [numeric_input_layer],
            outputs=output_layer,
        )
        return model

    def format_input_data(
        self,
        data: Union[None, pd.core.frame.DataFrame] = None,
        subset: Union[None, pd.core.series.Series] = None,
    ) -> List[Union[pd.core.series.Series, pd.core.frame.DataFrame]]:
        """List each categorical feature for input to own embedding layer."""
        if data is None:
            data = self.data
        subset = default_subset_to_all(subset, data)
        return split_categorical_features(
            data[subset], self.categorical_features, self.numeric_features
        )

    def train(
        self,
        params: Union[None, dict] = None,
        subset: Union[None, pd.core.series.Series] = None,
        validation_early_stopping: bool = True,
    ) -> keras.Model:
        """Train with survival loss function of Gensheimer, Narasimhan (2019).

        See https://peerj.com/articles/6257/.

        Use the AMSGrad variant of the Adam optimizer. See Reddi, Kale, and
        Kumar (2018) at https://openreview.net/forum?id=ryQu7f-RZ.

        Train until validation set performance does not improve for the given
        number of epochs or the given maximum number of epochs.

        Returns:
            A trained Keras model.
        """
        if params is None:
            params = {}
        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]
        train_subset = subset & ~self.data[self.validation_col]
        valid_subset = subset & self.data[self.validation_col]
        x_train = self.format_input_data(subset=train_subset)
        x_valid = self.format_input_data(subset=valid_subset)
        y_surv = make_surv_array(
            self.data[[self.duration_col, self.max_lead_col]].min(axis=1),
            self.data[self.event_col],
            np.arange(self.n_intervals + 1),
        )
        train_weights = data[train_subset][self.weight_col] if self.weight_col else None
        valid_weights = data[valid_subset][self.weight_col] if self.weight_col else None
        model = Model(self.model.inputs, self.model.outputs)
        model.compile(
            loss=surv_likelihood(self.n_intervals), optimizer=Adam(amsgrad=True)
        )
        callbacks = []
        if validation_early_stopping:
            callbacks.append(
                EarlyStopping(
                    monitor="val_loss",
                    mode="min",
                    verbose=1,
                    patience=self.config.get("PATIENCE", 4),
                    restore_best_weights=True,
                )
            )
        model.fit(
            x_train,
            y_surv[train_subset],
            batch_size=min(
                params.get("BATCH_SIZE", self.config.get("BATCH_SIZE", 512)),
                train_subset.sum(),
            ),
            epochs=params.get("MAX_EPOCHS", self.config.get("MAX_EPOCHS", 256)),
            validation_data=(x_valid, y_surv[valid_subset], valid_weights),
            verbose=2,
            callbacks=callbacks,
            sample_weight=train_weights,
        )
        return model

    def transform_features(self) -> pd.DataFrame:
        """Transform features to suit model training."""
        data = self.data.copy(deep=True)
        numeric_data = data.drop(self.reserved_cols, axis=1).select_dtypes(
            ["number", "datetime"]
        )
        minima = np.nanmin(numeric_data, axis=0)
        maxima = np.nanmax(numeric_data, axis=0)
        data[numeric_data.columns] = (numeric_data - minima) / (maxima - minima) - 0.5
        return data

    def predict(
        self,
        subset: Union[None, pd.core.series.Series] = None,
        custom_data: Union[None, pd.core.frame.DataFrame] = None,
        cumulative: bool = True,
    ) -> np.ndarray:
        """Use trained Keras model to predict observation survival rates.

        Args:
            subset: A Boolean Series that is True for observations for which
                predictions will be produced. If None, default to all
                observations.
            custom_data: A DataFrame in the same format as the input data for
                which predictions will be produced. If None, default to the
                assigned input data.
            cumulative: If True, produce cumulative survival probabilies.
                If False, produce marginal survival probabilities (i.e., one
                minus the hazard rate).

        Returns:
            A numpy array of survival probabilities by observation and lead
            length.
        """
        if cumulative:
            model = make_predictions_cumulative(self.model)
        else:
            model = make_predictions_marginal(self.model)
        return model.predict(self.format_input_data(data=custom_data, subset=subset))

    def save_model(self, file_name: str = "FFNN_Model", path: str = "") -> None:
        """Save the TensorFlow model to disk."""
        self.model.save(path + file_name + ".h5")

    def compute_shap_values(
        self, subset: Union[None, pd.core.series.Series] = None
    ) -> dict:
        """Compute SHAP values by lead length, observation, and feature.

        SHAP values for networks with embedding layers are not supported as of
        9 Jun 2020.

        Compute SHAP values for restricted mean survival time in addition to
        each lead length.

        Args:
            subset: A Boolean Series that is True for observations for which
                the shap values will be computed. If None, default to all
                observations.

        Returns:
            A dictionary of numpy arrays, each of which contains SHAP values
            for the outcome given by its key.
        """
        print(
            "SHAP values for networks with embedding layers are not "
            "supported as of 9 Jun 2020."
        )
        return None
        model = make_predictions_marginal(self.model)
        subset = default_subset_to_all(subset, self.data)
        shap_subset = split_categorical_features(
            self.data[subset], self.categorical_features, self.numeric_features
        )
        shap_subset = [np.atleast_2d(i.values) for i in shap_subset]
        interval_shap_values = shap.DeepExplainer(
            Model(model.inputs, model.output), shap_subset
        ).shap_values(shap_subset)
        model = make_predictions_cumulative(self.model)
        sum_layer = Lambda(lambda x: K.sum(x, axis=1))(model.output)
        rmst_shap_values = shap.DeepExplainer(
            Model(model.inputs, sum_layer), shap_subset
        ).shap_values(shap_subset)
        if self.categorical_features:
            interval_shap_values = [
                np.hstack(arr_list) for arr_list in interval_shap_values
            ]
            rmst_shap_values = np.hstack(rmst_shap_values)
        shap_values = {
            str(i + 1) + "_lead": arr for i, arr in enumerate(interval_shap_values)
        }
        shap_values["RMST"] = rmst_shap_values
        return shap_values

    def compute_model_uncertainty(
        self, subset: Union[None, pd.core.series.Series] = None, n_iterations: int = 200
    ) -> np.ndarray:
        """Predict with dropout as proposed by Gal and Ghahramani (2015).

        See https://arxiv.org/abs/1506.02142.

        Args:
            subset: A Boolean Series that is True for observations for which
                predictions will be produced. If None, default to all
                observations.
            n_iterations: Number of random dropout specifications to obtain
                predictions from.

        Returns:
            A numpy array of predictions by observation, lead length, and
            iteration.
        """
        subset = default_subset_to_all(subset, self.data)
        model_inputs = split_categorical_features(
            self.data[subset], self.categorical_features, self.numeric_features
        ) + [1.0]
        predict_with_dropout = K.function(
            self.model.inputs + [K.learning_phase()], self.model.outputs
        )
        predictions = np.dstack(
            [predict_with_dropout(model_inputs)[0] for i in range(n_iterations)]
        )
        return predictions


class TFSurvivalModeler(TFModeler, SurvivalModeler):
    """Use TensorFlow to forecast probabilities of being observed in future periods."""

    pass


class FeedforwardNeuralNetworkModeler(TFSurvivalModeler):
    """Deprecated alias for TFSurvivalModeler"""

    warn(
        "The name 'FeedforwardNeuralNetworkModeler' is deprecated. "
        "Please use 'TFSurvivalModeler' instead.",
        DeprecationWarning,
    )


FeedForwardNeuralNetworkModeler = FeedforwardNeuralNetworkModeler


class ProportionalHazardsModeler(TFSurvivalModeler):
    """Train a proportional hazards model with embeddings for categorical features using Keras."""

    def construct_embedding_network(self) -> keras.Model:
        """Set embedding layers that feed into a single node.

        Each categorical feature passes through its own embedding layer, which
        maps whole numbers to the real line.

        The embedded values and numeric features feed directly into a single node. The
        single node feeds into a proportional hazards layer.

        Returns:
            An untrained Keras model.
        """
        if self.categorical_features:
            categorical_input_layers = [
                Input(shape=(1,), dtype="int32") for _ in self.categorical_features
            ]
            embed_layers = [
                Embedding(
                    input_dim=self.data[col].max() + 2,
                    output_dim=int(
                        (self.data[col].nunique() + 2)
                        ** self.config.get("EMBED_EXPONENT", 0)
                    ),
                    embeddings_regularizer=l2(self.config.get("EMBED_L2_REG", 2.0)),
                )(lyr)
                for (col, lyr) in zip(
                    self.categorical_features, categorical_input_layers
                )
            ]
            flatten_layers = [Flatten()(lyr) for lyr in embed_layers]
            numeric_input_layer = Input(shape=(len(self.numeric_features),))
            concat_layer = Concatenate(axis=1)(flatten_layers + [numeric_input_layer])
            dense_layer = Dense(1, use_bias=0, kernel_initializer="zeros")(concat_layer)
        else:
            categorical_input_layers = []
            numeric_input_layer = Input(shape=(len(self.numeric_features),))
            dense_layer = Dense(1, use_bias=0, kernel_initializer="zeros")(
                numeric_input_layer
            )
        output_layer = PropHazards(self.n_intervals)(dense_layer)
        model = Model(
            inputs=categorical_input_layers + [numeric_input_layer],
            outputs=output_layer,
        )
        return model

    def save_model(self, file_name: str = "PH_Model", path: str = "") -> None:
        """Save the TensorFlow model to disk."""
        self.model.save(path + file_name + ".h5")

    def hyperoptimize(self, **kwargs) -> dict:
        """Returns None for ProportionalHazardsModeler, which does not have hyperparameters"""
        warn(
            "Warning: ProportionalHazardsModeler does not have hyperparameters to optimize."
        )
        return None


class ProportionalHazardsEncodingModeler(TFSurvivalModeler):
    """Train a proportional hazards model with binary-encoded categorical features using Keras."""

    def build_model(self, n_intervals: Union[None, int] = None) -> None:
        """Train and store a neural network with a proportional hazards restriction."""
        tf.random.set_seed(self.config.get("SEED", 9999))
        if n_intervals:
            self.n_intervals = n_intervals
        else:
            self.n_intervals = self.set_n_intervals()
        for col in self.categorical_features:
            binary_encodings = binary_encode_feature(self.data[col])
            del self.data[col]
            self.data[binary_encodings.columns] = binary_encodings
            del binary_encodings
        self.data = self.data.fillna(self.config.get("NON_CAT_MISSING_VALUE", -1))
        self.model = self.construct_network()
        self.model = self.train()
        self.model = make_predictions_cumulative(self.model)

    def format_input_data(
        self,
        data: Union[None, pd.core.frame.DataFrame] = None,
        subset: Union[None, pd.core.series.Series] = None,
    ) -> List[Union[pd.core.series.Series, pd.core.frame.DataFrame]]:
        """Keep only the features and observations desired for model input."""
        if data is None:
            data = self.data
        subset = default_subset_to_all(subset, data)
        formatted_data = data.drop(self.reserved_cols, axis=1)[subset]
        for col in self.categorical_features:
            formatted_data[col] = formatted_data[col].cat.codes
        return formatted_data

    def construct_network(self) -> keras.Model:
        """Set all features to feed directly into a single node.

        The single node feeds into a proportional hazards layer.

        Returns:
            An untrained Keras model.
        """
        input_layer = Input(
            shape=(len(self.data.drop(self.reserved_cols, axis=1).columns),)
        )
        dense_layer = Dense(1, use_bias=0, kernel_initializer="zeros")(input_layer)
        output_layer = PropHazards(self.n_intervals)(dense_layer)
        model = Model(inputs=input_layer, outputs=output_layer)
        return model

    def save_model(self, file_name: str = "PH_Encoded_Model", path: str = "") -> None:
        """Save the TensorFlow model to disk."""
        self.model.save(path + file_name + ".h5")

    def hyperoptimize(self, **kwargs) -> dict:
        """Returns None for ProportionalHazardsEncodingModeler, which does not have hyperparameters"""
        warn(
            "Warning: ProportionalHazardsEncodingModeler does not have hyperparameters to optimize."
        )
        return None



#!/usr/bin/env python3
"""Execute default FIFE pipeline.

This script trains a discrete-time survival model on unbalanced panel data,
then forecasts survival for the observations in the most recent period.
It creates and populates Intermediate and Output folders in the given
directory.

See the accompanying README file for further details and instructions,
including documentation of all configuration parameters.
"""

import json
import os
from time import time
from typing import Type

from fife import lgb_modelers, processors, tf_modelers, utils
from fife.base_modelers import Modeler
import numpy as np
import pandas as pd
from sklearn.calibration import calibration_curve


def main():
    """Execute default FIFE pipeline from data to survival forecasts and metrics."""
    checkpoint_time = time()
    config = parse_config()
    if config.get("EXIT_COL_PATH"):
        raise NotImplementedError(
            "Forecasting exit circumstances from the command line is not yet supported. Try LGBExitModeler from the FIFE Python package."
        )
    if config.get("STATE_COL"):
        raise NotImplementedError(
            "Forecasting future state from the command line is not yet supported. Try LGBStateModeler from the FIFE Python package."
        )
    utils.make_results_reproducible(config["SEED"])
    utils.redirect_output_to_log(path=config["RESULTS_PATH"])
    utils.print_copyright()
    utils.print_config(config)
    data = read_data(config)
    print(f"I/O setup time: {time() - checkpoint_time} seconds")

    checkpoint_time = time()
    data_processor = processors.PanelDataProcessor(config, data)
    data_processor.build_processed_data()
    utils.save_intermediate_data(
        data_processor.data,
        "Processed_Data",
        file_format="pickle",
        path=config["RESULTS_PATH"],
    )
    print(f"Data processing time: {time() - checkpoint_time} seconds")

    by_feature = config.get("BY_FEATURE", "")
    if by_feature != "" and by_feature not in data.columns:
        raise ValueError(
            "The selected feature for 'BY_FEATURE' is not in the dataset. Check spelling or the original dataset to ensure that you are entering the correct feature name."
        )

    checkpoint_time = time()
    utils.ensure_folder_existence(f'{config["RESULTS_PATH"]}/Intermediate/Models')
    test_intervals = config.get("TEST_INTERVALS", config.get("TEST_PERIODS", 0) - 1)
    if config.get("TREE_MODELS"):
        modeler_class = lgb_modelers.LGBSurvivalModeler
    elif config.get("PROPORTIONAL_HAZARDS"):
        modeler_class = tf_modelers.ProportionalHazardsModeler
    else:
        modeler_class = tf_modelers.FeedforwardNeuralNetworkModeler
    modeler = modeler_class(config=config, data=data_processor.data)
    modeler.n_intervals = (
        test_intervals if test_intervals > 0 else modeler.set_n_intervals()
    )
    if not config.get("TIME_ID_AS_FEATURE"):
        modeler.numeric_features.remove(config["TIME_IDENTIFIER"])
    if config.get("HYPER_TRIALS", 0) > 0:
        params = modeler.hyperoptimize(config["HYPER_TRIALS"])
    else:
        params = None
    modeler.build_model(n_intervals=modeler.n_intervals, params=params)
    modeler.save_model(path=f"{config['RESULTS_PATH']}/Intermediate/Models/")
    print(f"Model training time: {time() - checkpoint_time} seconds")

    checkpoint_time = time()
    if test_intervals > 0:

        # Save metrics
        max_test_intervals = int((len(set(modeler.data["_period"])) - 1) / 2)

        evaluation_subset = modeler.data["_period"] == (
            modeler.data["_period"].max() - min(test_intervals, max_test_intervals)
        )

        if by_feature != "":
            values = list(set(data[by_feature]))

            for feature_value in values:

                evaluation_subset_by_feature = modeler.data[by_feature] == feature_value
                evaluation_subset_comparison = (
                    evaluation_subset & evaluation_subset_by_feature
                )

                utils.save_output_table(
                    modeler.evaluate(evaluation_subset_comparison),
                    f"Metrics_{feature_value}",
                    path=config["RESULTS_PATH"],
                )

        utils.save_output_table(
            modeler.evaluate(evaluation_subset),
            "Metrics",
            path=config["RESULTS_PATH"],
        )

        # Save counts by quantile
        utils.save_output_table(
            modeler.tabulate_survival_by_quantile(
                n_quantiles=config["QUANTILES"],
                subset=evaluation_subset,
            ),
            "Counts_by_Quantile",
            index=False,
            path=config["RESULTS_PATH"],
        )

        # Save forecast errors
        actuals = np.array(
            [
                modeler.data[evaluation_subset]["_duration"] > time_horizon
                for time_horizon in range(test_intervals)
            ]
        ).T
        predictions = modeler.predict(evaluation_subset)
        utils.save_output_table(
            pd.DataFrame(
                predictions - actuals,
                columns=[
                    f"{time_horizon + 1}-period Forecast Error"
                    for time_horizon in range(test_intervals)
                ],
            ),
            "Forecast_Errors",
            index=False,
            path=config["RESULTS_PATH"],
        )

        # Save calibration errors
        predicted_share, actual_share = calibration_curve(
            actuals.flatten(), predictions.flatten(), n_bins=8, strategy="quantile"
        )
        calibration_errors = pd.DataFrame(
            [predicted_share, actual_share, actual_share - predicted_share]
        ).T
        calibration_errors.columns = [
            "Predicted Share",
            "Actual Share",
            "Calibration Error",
        ]
        calibration_errors.index.name = "Quantile"
        calibration_errors.index = calibration_errors.index + 1
        utils.save_output_table(
            calibration_errors,
            "Calibration_Errors",
            path=config["RESULTS_PATH"],
        )

    else:

        # Save forecasts
        individual_predictions = modeler.forecast()
        utils.save_output_table(
            individual_predictions, "Survival_Curves", path=config["RESULTS_PATH"]
        )

        # Save aggregated forecasts with uncertainty intervals
        utils.save_output_table(
            utils.compute_aggregation_uncertainty(individual_predictions),
            "Aggregate_Survival_Bounds",
            index=False,
            path=config["RESULTS_PATH"],
        )

        # Save and plot actual, fitted, and forecasted retention rates
        lead_periods = config["RETENTION_INTERVAL"]
        time_ids = pd.factorize(
            modeler.data[modeler.config["TIME_IDENTIFIER"]], sort=True
        )[0]
        retention_rates = modeler.tabulate_retention_rates(
            lead_periods=lead_periods, time_ids=time_ids
        )
        utils.save_output_table(
            retention_rates, "Retention_Rates", path=config["RESULTS_PATH"]
        )
        axes = retention_rates.plot()
        axes.set_ylabel(f"{lead_periods}-period Retention Rate")
        earliest_period = data_processor.data[
            data_processor.config["TIME_IDENTIFIER"]
        ].min()
        axes.set_xlabel(f"Periods Since {earliest_period}")
        utils.save_plot("Retention_Rates", path=config["RESULTS_PATH"])

        # Plot SHAP values for a subset of observations in the final period
        sample_size = config.get("SHAP_SAMPLE_SIZE", 0)
        if (
            isinstance(modeler, (lgb_modelers.GradientBoostedTreesModeler))
            and sample_size > 0
        ):
            shap_observations = (
                modeler.data[modeler.data["_predict_obs"]]
                .sample(n=sample_size)
                .sort_index()
            )
            subset = modeler.data.index.isin(shap_observations.index)
            shap_values = modeler.compute_shap_values(subset=subset)
            utils.plot_shap_values(
                shap_values,
                shap_observations[
                    modeler.categorical_features + modeler.numeric_features
                ],
                modeler.data[subset][
                    modeler.categorical_features + modeler.numeric_features
                ],
                config["TIME_IDENTIFIER"],
                path=config["RESULTS_PATH"],
            )

    print(f"Output production time: {time() - checkpoint_time} seconds")


def parse_config() -> dict:
    """Parse configuration parameters specified in the command line."""
    parser = utils.FIFEArgParser()
    args = parser.parse_args()
    config = {}
    if args.CONFIG_PATH:
        with open(args.CONFIG_PATH, "r") as file:
            config.update(json.load(file))
    config.update({k: v for k, v in vars(args).items() if v is not None})
    return config


def read_data(config: dict) -> pd.DataFrame:
    """Read the input dataset as specified in config or inferred from current directory."""
    if "DATA_FILE_PATH" not in config.keys():
        valid_suffixes = (".csv", ".csv.gz", ".p", ".pkl", ".h5")
        candidate_data_files = [
            file for file in os.listdir() if file.endswith(valid_suffixes)
        ]
        assert len(candidate_data_files) >= 1, (
            "No data files found in current directory. "
            f"Valid data file suffixes are {valid_suffixes}. "
            "If you want to use data in another directory, "
            "please specify the DATA_FILE_PATH."
        )
        assert len(candidate_data_files) <= 1, (
            "Multiple data files found in current directory. "
            "please specify the DATA_FILE_PATH."
        )
        print(f"Using {candidate_data_files[0]} as data file.")
        config["DATA_FILE_PATH"] = candidate_data_files[0]
    data = utils.import_data_file(config["DATA_FILE_PATH"])
    return data


if __name__ == "__main__":
    main()



"""I/O, logging, calculation, and plotting functions for FIFE."""

import argparse
from distutils.util import strtobool
import json
import os
import pickle
import random as rn
import sys
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import shap


def make_results_reproducible(seed: int = 9999) -> None:
    """Ensure executing from a fresh state produces identical results."""
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    rn.seed(seed)


def ensure_folder_existence(path: str = "") -> None:
    """Create a directory if it doesn't already exist."""
    if not os.path.exists(path):
        os.makedirs(path)


def redirect_output_to_log(path: str = "") -> None:
    """Send future output to a text file instead of console."""
    ensure_folder_existence(f"{path}/Output/Logs/")
    sys.stdout = open(f"{path}/Output/Logs/Log.txt", "w")


def print_config(config: dict) -> None:
    """Neatly print given dictionary of config parameters."""
    title = "CONFIGURATION PARAMETERS"
    print(title)
    print("-" * len(title))
    print(json.dumps(config, indent=4) + "\n")


def print_copyright() -> None:
    print("Produced using FIFE: Finite-Interval Forecasting Engine")
    print("Copyright (c) 2018 - 2020, Institute for Defense Analyses (IDA)")
    print("Please cite using the suggested citation in the LICENSE file.\n")


def import_data_file(file_path: str = "Input_Data") -> pd.core.frame.DataFrame:
    """Return the data stored in given file in given folder."""
    if file_path.endswith(".feather"):
        data = pd.read_feather(file_path)
    elif file_path.endswith(".csv"):
        data = pd.read_csv(file_path)
    elif file_path.endswith(".csv.gz"):
        data = pd.read_csv(file_path)
    elif file_path.endswith((".p", "pkl")):
        data = pd.read_pickle(file_path)
    elif file_path.endswith(".h5"):
        data = pd.read_hdf(file_path)
    elif file_path.endswith(".json"):
        data = pd.read_json(file_path)
    else:
        raise Exception(
            "Data file extension is invalid. See README for " + "valid file extensions."
        )
    return data


def compute_aggregation_uncertainty(
    individual_probabilities: pd.core.frame.DataFrame, percent_confidence: float = 0.95
) -> pd.core.frame.DataFrame:
    """Statistically bound number of events given each of their probabilities.

    Args:
        individual_probabilities: A DataFrame of probabilities where each row
            represents an individual and each column represents an event.
        percent_confidence: The percent confidence of the two-sided intervals
            defined by the computed bounds.

    Raises:
        ValueError: If percent_confidence is outside of the interval (0, 1).

    Returns:
        A DataFrame containing, for each column in individual_probabilities,
        the expected number of events and interval bounds on the number of
        events.
    """
    if (percent_confidence <= 0) | (percent_confidence >= 1):
        raise ValueError("Percent confidence outside (0, 1)")
    means = individual_probabilities.transpose().to_numpy().sum(axis=1)
    cutoff = np.log((1 - percent_confidence) / 2) / means
    upper_bounds = (
        1 + ((-cutoff) + np.sqrt(np.square(cutoff) - (8 * cutoff))) / 2
    ) * means
    upper_bounds = np.minimum(upper_bounds, individual_probabilities.shape[0])
    lower_bounds = np.maximum((1 - np.sqrt(-2 * cutoff)) * means, 0)
    lower_bounds = np.maximum(lower_bounds, 0)
    conf = str(int(percent_confidence * 100))
    times = list(individual_probabilities)
    output = pd.DataFrame(
        {
            "Event": times,
            "ExpectedEventCount": means,
            "Lower" + conf + "PercentBound": lower_bounds,
            "Upper" + conf + "PercentBound": upper_bounds,
        }
    )
    return output


def plot_binary_prediction_errors(
    errors: dict,
    width: float = 8,
    height: float = 1,
    alpha: float = 2 ** -8,
    color: str = "black",
    center_tick_color: str = "green",
    center_tick_height: float = 0.125,
    path: str = "",
) -> None:
    """Make a rug plot of binary prediction errors.

    Args:
        errors: A dictionary of numpy arrays, each of which contains error values
            for the outcome given by its key.
        width: Width of the rug plot.
        height: Height of the rug plot.
        alpha: The opacity of plotted ticks, from 2e-8 (nearly transparent) to
            1 (opaque).
        color: The color of plotted ticks.
        center_tick_color: The color of the ticks marking the center of the plot.
        center_tick_height: The height of the ticks marking the center of the plot.
        path: The path preceding the Output folder in which the plots will be
            saved.
    """
    sns.set(rc={"figure.figsize": (width, height)})
    for key, arr in errors.items():
        sns.rugplot(arr, height=height, alpha=alpha, color=color)
        plt.xlim(-1, 1)
        plt.axvline(ymin=0, ymax=center_tick_height, color=center_tick_color)
        plt.axvline(
            ymin=height - center_tick_height, ymax=height, color=center_tick_color
        )
        plt.xticks([])
        plt.yticks([])
        save_plot(f"Errors_{key}", path=path)


def plot_shap_values(
    shap_values: dict,
    raw_data: pd.core.frame.DataFrame,
    processed_data: Union[None, pd.core.frame.DataFrame] = None,
    no_summary_col: str = Union[None, str],
    alpha: float = 0.5,
    path: str = "",
) -> None:
    """Make plots of SHAP values.

    SHAP values quantify feature contributions to predictions.

    Args:
        shap_values: A dictionary of numpy arrays, each of which contains SHAP
            values for the outcome given by its key.
        raw_data: Feature values prior to processing into model input.
        processed_data: Feature values used as model input.
        no_summary_col: The name of a column to never use for summary plots.
        alpha: The opacity of plotted points, from 2e-8 (nearly transparent) to
            1 (opaque).
        path: The path preceding the Output folder in which the plots will be
            saved.
    """
    shap.initjs()
    if processed_data is None:
        processed_data = raw_data
    for col in processed_data.select_dtypes("category"):
        processed_data[col] = processed_data[col].cat.codes
    for key, arr in shap_values.items():
        shap.summary_plot(
            arr, plot_type="bar", feature_names=raw_data.columns, show=False
        )
        save_plot(f"Importance_{key}", path=path)
        shap.summary_plot(arr, raw_data, alpha=alpha, show=False)
        save_plot(f"Summary_{key}", path=path)
        if raw_data.columns[np.argmax(np.abs(arr).mean(axis=0))] == no_summary_col:
            shap.dependence_plot(
                f"rank(1)",
                arr,
                processed_data,
                display_features=raw_data,
                alpha=alpha,
                show=False,
            )
        else:
            shap.dependence_plot(
                f"rank(0)",
                arr,
                processed_data,
                display_features=raw_data,
                alpha=alpha,
                show=False,
            )
        save_plot(f"Dependence_{key}", path=path)


def save_maps(
    obj: Union[pd.core.series.Series, dict], file_name: str, path: str = ""
) -> None:
    """Save a map from values to other values to Intermediate folder."""
    ensure_folder_existence(f"{path}/Intermediate/Data")
    with open(f"{path}/Intermediate/Data/{file_name}.p", "wb") as file:
        pickle.dump(obj, file)


def save_intermediate_data(
    data: pd.core.frame.DataFrame,
    file_name: str,
    file_format: str = "pickle",
    path: str = "",
) -> None:
    """Save given DataFrame in Intermediate folder in given format."""
    ensure_folder_existence(f"{path}/Intermediate/Data")
    if file_format == "csv":
        data.to_csv(f"{path}/Intermediate/Data/{file_name}.csv")
    elif file_format == "feather":
        data.to_feather(f"{path}/Intermediate/Data/{file_name}.feather")
    elif file_format == "hdf":
        data.to_hdf(f"{path}/Intermediate/Data/{file_name}.h5", key=file_name)
    elif file_format == "pickle":
        data.to_pickle(f"{path}/Intermediate/Data/{file_name}.p")
    else:
        raise Exception(
            "File format invalid. Valid file formats "
            + 'are "csv", "feather", "hdf", and "pickle".'
        )


def save_output_table(
    data: pd.core.frame.DataFrame, file_name: str, index: bool = True, path: str = ""
) -> None:
    """Save given DataFrame in the Output folder as a csv file."""
    ensure_folder_existence(f"{path}/Output/Tables")
    data.to_csv(f"{path}/Output/Tables/{file_name}.csv", index=index)


def save_plot(file_name: str, path: str = "") -> None:
    """Save the most recently plotted plot in high resolution."""
    ensure_folder_existence(f"{path}/Output/Figures")
    plt.tight_layout()
    plt.savefig(f"{path}/Output/Figures/{file_name}.png", dpi=512)
    plt.close()


def create_example_data(
    n_persons: int = 8192, n_periods: int = 12
) -> pd.core.frame.DataFrame:
    """Fabricate an unbalanced panel dataset suitable as FIFE input."""
    seed = 9999
    np.random.seed(seed)
    values = []
    for i in np.arange(n_persons):
        period = np.random.randint(n_periods) + 1
        x_1 = np.random.uniform()
        x_2 = rn.choice(["A", "B", "C"])
        x_3 = np.random.uniform() + 1.0
        x_4_categories = ["a", "b", "c", 1, 2, 3, np.nan]
        x_4 = rn.choice(x_4_categories)
        while period <= n_periods:
            values.append([i, period, x_1, x_2, x_3, x_4])
            if x_2 == "A":
                x_1 += np.random.uniform(0, 0.1)
            else:
                x_1 += np.random.uniform(0, 0.2)
            if x_1 > np.sqrt(x_3):
                break
            if x_4 in x_4_categories[:-2]:
                x_4_transition_value = x_4_categories[x_4_categories.index(x_4) + 1]
                if np.random.uniform() >= 0.75:
                    x_4 = x_4_transition_value
                    del x_4_transition_value
            period += 1
    values = pd.DataFrame(
        values,
        columns=[
            "individual",
            "period",
            "feature_1",
            "feature_2",
            "feature_3",
            "feature_4",
        ],
    )
    values["feature_5"] = values["feature_2"]
    return values


class FIFEArgParser(argparse.ArgumentParser):
    """Argument parser for the FIFE command-line interface."""

    def __init__(self):
        super().__init__()
        self.add_argument(
            "CONFIG_PATH",
            type=str,
            nargs="?",
            default=None,
            help="A relative or absolute path to a configuration file.",
        )
        self.add_argument(
            "--DATA_FILE_PATH",
            type=str,
            help="A relative or absolute path to the input data file.",
        )
        self.add_argument(
            "--NOTES_FOR_LOG",
            type=str,
            help="Custom text that will be printed in the log produced during execution.",
        )
        self.add_argument(
            "--RESULTS_PATH",
            type=str,
            default="FIFE_results",
            help="The relative or absolute path on which Intermediate and Output folders will be stored.",
        )
        self.add_argument(
            "--SEED",
            type=int,
            default=9999,
            help="The initializing value for all random number generators.",
        )
        self.add_argument(
            "--INDIVIDUAL_IDENTIFIER",
            type=str,
            help="The name of the feature that identifies individuals that persist over multiple time periods in the data.",
        )
        self.add_argument(
            "--TIME_IDENTIFIER",
            type=str,
            help="The name of the feature that identifies time periods in the data.",
        )
        self.add_argument(
            "--BY_FEATURE",
            type=str,
            help="The name of the feature by which to produce separate Metrics.csv files.",
        )
        self.add_argument(
            "--CATEGORICAL_SUFFIXES",
            type=str,
            nargs="+",
            help="Optional list of suffixes denoting that columns ending with such a suffix should be treated as categorical.",
        )
        self.add_argument(
            "--DATETIME_AS_DATE",
            type=bool,
            help="How datetime features will be represented for the gradient-boosted trees modeler.",
        )
        self.add_argument(
            "--MAX_NULL_SHARE",
            type=float,
            help="The maximum share of observations that may have a null value for a feature to be kept for training.",
        )
        self.add_argument(
            "--MAX_UNIQUE_NUMERIC_CATS",
            type=int,
            help="The maximum number of unique values for a feature of a numeric type to be considered categorical.",
        )
        self.add_argument(
            "--NUMERIC_SUFFIXES",
            type=str,
            nargs="+",
            help="Optional list of suffixes denoting that columns ending with such a suffix should be treated as numeric.",
        )
        self.add_argument(
            "--TIME_ID_AS_FEATURE",
            type=bool,
            help="Whether or not the time identifier will be included as a feature.",
        )
        self.add_argument(
            "--MIN_SURVIVORS_IN_TRAIN",
            type=int,
            help="The minimum number of training set observations surviving a given time horizon for the model to be trained to make predictions for that time horizon.",
        )
        self.add_argument(
            "--TEST_INTERVALS",
            type=int,
            help="The number of most recent periods to treat as absent from the data during training for the purpose of model evaluation.",
        )
        self.add_argument(
            "--TEST_PERIODS",
            type=int,
            help="One plus the value represented by TEST_INTERVALS. Deprecated and overriden by TEST_INTERVALS.",
        )
        self.add_argument(
            "--VALIDATION_SHARE",
            type=float,
            help="The share of observations used for evaluation instead of training for hyperoptimization or early stopping.",
        )
        self.add_argument(
            "--EXIT_FILE_PATH",
            type=str,
            help="The path to a file identifying the circumstances of exit for each individual observed to exit in the next period. If given, FIFE will forecast the circumstances of exit conditional on exit instead of forecasting survival.",
        )
        self.add_argument(
            "--STATE_COL",
            type=str,
            help="The name of the column representing the future state to forecast. If given and EXIT_FILE_PATH not given, FIFE will forecast the future state conditional on survival instead of forecasting survival.",
        )
        self.add_argument(
            "--TREE_MODELS",
            type=strtobool,
            default=True,
            help="Whether FIFE will train gradient-boosted trees, as opposed to a neural network.",
        )
        self.add_argument(
            "--HYPER_TRIALS",
            type=int,
            help="The number of hyperparameter sets to trial.",
        )
        self.add_argument(
            "--MAX_EPOCHS",
            type=int,
            help="If HYPER_TRIALS is zero, the maximum number of passes through the training set.",
        )
        self.add_argument(
            "--PATIENCE",
            type=int,
            help="If HYPER_TRIALS is zero, the number of passes through the training dataset without improvement in validation set performance before training is stopped early.",
        )
        self.add_argument(
            "--BATCH_SIZE",
            type=int,
            help="The number of observations per batch of data fed to the machine learning model for training. ",
        )
        self.add_argument(
            "--DENSE_LAYERS",
            type=int,
            help="The number of dense layers in the neural network.",
        )
        self.add_argument(
            "--DROPOUT_SHARE",
            type=float,
            help="The probability of a densely connected node of the neural network being set to zero weight during training.",
        )
        self.add_argument(
            "--EMBED_EXPONENT",
            type=float,
            help="	The ratio of the natural logarithm of the number of embedded values to the natural logarithm of the number of unique categories for each categorical feature.",
        )
        self.add_argument(
            "--EMBED_L2_REG",
            type=float,
            help="The L2 regularization coefficient for each embedding layer.",
        )
        self.add_argument(
            "--NODES_PER_DENSE_LAYER",
            type=int,
            help="The number of nodes per dense layer in the neural network.",
        )
        self.add_argument(
            "--NON_CAT_MISSING_VALUE",
            type=float,
            help="The value used to replace missing values of numeric features.",
        )
        self.add_argument(
            "--PROPORTIONAL_HAZARDS",
            type=bool,
            help="Whether FIFE will restrict the neural network to a proportional hazards model.",
        )
        self.add_argument(
            "--QUANTILES",
            type=int,
            default=5,
            help="The number of similarly-sized bins for which survival and total counts will be reported for each time horizon.",
        )
        self.add_argument(
            "--RETENTION_INTERVAL",
            type=int,
            default=1,
            help="The number of periods over which retention rates are computed.",
        )
        self.add_argument(
            "--SHAP_PLOT_ALPHA",
            type=float,
            help="The transparency of points in SHAP plots.",
        )
        self.add_argument(
            "--SHAP_SAMPLE_SIZE",
            type=int,
            default=128,
            help="The number of observations randomly sampled for SHAP value calculation and plotting.",
        )



"""The abstract classes on which all FIFE modelers are based."""

from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Union

from lifelines.utils import concordance_index
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix, r2_score, roc_auc_score


def default_subset_to_all(
    subset: Union[None, pd.core.series.Series], data: pd.core.frame.DataFrame
) -> pd.core.series.Series:
    """Map an unspecified subset to an entirely True boolean mask."""
    if subset is None:
        return pd.Series([True] * data.shape[0], index=data.index)
    return subset


def compute_metrics_for_binary_outcome(
    actuals: Union[pd.Series, pd.DataFrame],
    predictions: np.ndarray,
    threshold_positive: Union[None, str, float] = 0.5,
    share_positive: Union[None, str, float] = None,
    weights: Union[None, np.ndarray] = None,
) -> OrderedDict:
    """Evaluate predicted probabilities against actual binary outcome values.

    Args:
        actuals: A Series representing actual Boolean outcome values.
        predictions: A Series of predicted probabilities of the respective
            outcome values.
        threshold_positive: None, "predicted", or a value in [0, 1]
            representing the minimum predicted probability considered to be a
            positive prediction; Specify "predicted" to use the predicted share
            positive in each time horizon. Overridden by share_positive.
        share_positive: None, "predicted", or a value in [0, 1] representing
            the share of observations with the highest predicted probabilities
            considered to have positive predictions. Specify "predicted" to use
            the predicted share positive in each time horizon. Probability ties
            may cause share of positives in output to exceed given value.
            Overrides threshold_positive.
        weights: A 1-D array of weights with the same length as actuals and predictions.
        Each prediction contributes to metrics in proportion to its weight. If None, each prediction has the same weight.

    Returns:
        An ordered dictionary containing key-value pairs for area under the
        receiver operating characteristic curve (AUROC), predicted and
        actual shares of observations with an outcome of True, and all
        elements of the confusion matrix.
    """
    metrics = OrderedDict()
    if actuals.any() and not actuals.all():
        metrics["AUROC"] = roc_auc_score(actuals, predictions, sample_weight=weights)
    else:
        metrics["AUROC"] = np.nan
    metrics["Predicted Share"] = np.average(predictions, weights=weights)
    metrics["Actual Share"] = np.average(actuals, weights=weights)
    if actuals.empty:
        (
            metrics["True Positives"],
            metrics["False Negatives"],
            metrics["False Positives"],
            metrics["True Negatives"],
        ) = [0, 0, 0, 0]
    else:
        if share_positive == "predicted":
            share_positive = metrics["Predicted Share"]
        if share_positive is not None:
            threshold_positive = np.quantile(predictions, 1 - share_positive)
        elif threshold_positive == "predicted":
            threshold_positive = metrics["Predicted Share"]
        (
            metrics["True Positives"],
            metrics["False Negatives"],
            metrics["False Positives"],
            metrics["True Negatives"],
        ) = (
            confusion_matrix(
                actuals,
                predictions >= threshold_positive,
                labels=[True, False],
                sample_weight=weights,
            )
            .ravel()
            .tolist()
        )
    return metrics


def compute_metrics_for_categorical_outcome(
    actuals: pd.DataFrame,
    predictions: np.ndarray,
    weights: Union[None, np.ndarray] = None,
) -> OrderedDict:
    """Evaluate predicted probabilities against actual categorical outcome values.

    Args:
        actuals: A DataFrame representing one-hot-encoded actual class membership.
        predictions: A DataFrame of predicted probabilities of class membership
            for the respective observations represented in actuals.
        weights: A 1-D array of weights with the same length as actuals and predictions.
        Each prediction contributes to metrics in proportion to its weight. If None, each prediction has the same weight.

    Returns:
        An ordered dictionary containing a key-value pair for area under the
        receiver operating characteristic curve (AUROC). Classes with no positive values are not included in AUROC calculation.
    """
    metrics = OrderedDict()
    if actuals.all().any():
        metrics["AUROC"] = np.nan
    else:
        positive_cols = actuals.any()
        metrics["AUROC"] = roc_auc_score(
            actuals.loc[:, positive_cols],
            predictions[:, positive_cols],
            multi_class="ovr",
            average="weighted",
            sample_weight=weights,
        )
    return metrics


def compute_metrics_for_numeric_outcome(
    actuals: pd.core.series.Series,
    predictions: np.ndarray,
    weights: Union[None, np.ndarray] = None,
) -> OrderedDict:
    """Evaluate predicted numeric values against actual outcome values.

    Args:
        actuals: A Series representing actual outcome values.
        predictions: A Series of predictions for the respective outcome values.

    Returns:
        An ordered dictionary containing a key-value pair for R-squared.
    """
    metrics = OrderedDict()
    non_null_obs = actuals.notnull()
    metrics["R-squared"] = r2_score(
        actuals[non_null_obs],
        predictions[non_null_obs],
        sample_weight=weights[non_null_obs] if weights else None,
    )
    return metrics


class Modeler(ABC):
    """Set template for modelers that use panel data to produce forecasts.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast
        allow_gaps (bool): Whether or not observations should be included for
            training and evaluation if there is a period without an observation
            between the period of observations and the last period of the
            given time horizon.
    """

    def __init__(
        self,
        config: Union[None, dict] = {},
        data: Union[None, pd.core.frame.DataFrame] = None,
        duration_col: str = "_duration",
        event_col: str = "_event_observed",
        predict_col: str = "_predict_obs",
        test_col: str = "_test",
        validation_col: str = "_validation",
        period_col: str = "_period",
        max_lead_col: str = "_maximum_lead",
        spell_col: str = "_spell",
        weight_col: Union[None, str] = None,
        allow_gaps: bool = False,
    ) -> None:
        """Characterize data for modelling.

        Identify each column as categorical, reserved, and numeric.
        Compute the maximum lead length for modelling.

        Args:
            config: User-provided configuration parameters.
            data: User-provided panel data.
            duration_col: Name of the column representing the number of future
                periods observed for the given individual.
            event_col: Name of the column indicating whether the individual is
                observed to exit the dataset.
            predict_col: Name of the column indicating whether the observation
                will be used for prediction after training.
            test_col: Name of the column indicating whether the observation
                will be used for testing model performance after training.
            validation_col: Name of the column indicating whether the
                observation will be used for evaluating model performance
                during training.
            period_col (str): Name of the column representing the number of
                periods since the earliest period in the data.
            max_lead_col (str): Name of the column representing the number of
                observable future periods.
            spell_col (str): Name of the column representing the number of
                previous spells of consecutive observations of the same individual.
            weight_col (str): Name of the column representing observation weights.
            allow_gaps (bool): Whether or not observations should be included for
                training and evaluation if there is a period without an observation
                between the period of observations and the last period of the
                given time horizon.
        """
        if (config.get("TIME_IDENTIFIER", "") == "") and data is not None:
            config["TIME_IDENTIFIER"] = data.columns[1]
            print(
                "Time identifier column name not given; assumed to be "
                f'second-leftmost column ({config["TIME_IDENTIFIER"]})'
            )
        if (config.get("INDIVIDUAL_IDENTIFIER", "") == "") and data is not None:
            config["INDIVIDUAL_IDENTIFIER"] = data.columns[0]
            print(
                "Individual identifier column name not given; assumed to be "
                f'leftmost column ({config["INDIVIDUAL_IDENTIFIER"]})'
            )
        self.config = config
        self.data = data
        self.duration_col = duration_col
        self.event_col = event_col
        self.predict_col = predict_col
        self.test_col = test_col
        self.validation_col = validation_col
        self.period_col = period_col
        self.max_lead_col = max_lead_col
        self.spell_col = spell_col
        self.weight_col = weight_col
        self.allow_gaps = allow_gaps
        self.reserved_cols = [
            self.duration_col,
            self.event_col,
            self.predict_col,
            self.test_col,
            self.validation_col,
            self.period_col,
            self.max_lead_col,
            self.spell_col,
        ]
        if self.config:
            self.reserved_cols.append(self.config["INDIVIDUAL_IDENTIFIER"])
        if self.weight_col:
            self.reserved_cols.append(self.weight_col)
        if self.data is not None:
            self.categorical_features = [
                col for col in self.data.select_dtypes("category")
            ]
            self.numeric_features = [
                feature
                for feature in self.data
                if feature not in (self.categorical_features + self.reserved_cols)
            ]
            self.data = self.transform_features()

    @abstractmethod
    def train(self) -> Any:
        """Train and return a model."""

    @abstractmethod
    def predict(
        self, subset: Union[None, pd.core.series.Series] = None, cumulative: bool = True
    ) -> np.ndarray:
        """Use trained model to produce observation survival probabilities."""

    @abstractmethod
    def evaluate(
        self,
        subset: Union[None, pd.core.series.Series] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:
        """Tabulate model performance metrics."""

    @abstractmethod
    def forecast(self) -> pd.core.frame.DataFrame:
        """Tabulate survival probabilities for most recent observations."""

    @abstractmethod
    def subset_for_training_horizon(
        self, data: pd.DataFrame, time_horizon: int
    ) -> pd.DataFrame:
        """Return only observations where the outcome is observed."""

    @abstractmethod
    def label_data(self, time_horizon: int) -> pd.DataFrame:
        """Return data with an outcome label for each observation."""

    @abstractmethod
    def save_model(self, path: str = "") -> None:
        """Save model file(s) to disk."""

    def transform_features(self) -> pd.DataFrame:
        """Transform datetime features to suit model training."""
        return self.data

    def build_model(self, n_intervals: Union[None, int] = None) -> None:
        """Configure, train, and store a model."""
        if n_intervals:
            self.n_intervals = n_intervals
        else:
            self.n_intervals = self.set_n_intervals()
        self.model = self.train()

    def set_n_intervals(self) -> int:
        """Determine the maximum periods ahead the model will predict."""
        train_durations = self.data[[self.duration_col, self.max_lead_col]].min(axis=1)
        train_obs_by_lead_length = train_durations[
            (
                ~self.data[self.validation_col]
                & ~self.data[self.test_col]
                & ~self.data[self.predict_col]
            )
        ].value_counts()
        n_intervals = train_obs_by_lead_length[
            train_obs_by_lead_length > self.config.get("MIN_SURVIVORS_IN_TRAIN", 64)
        ].index.max()
        return n_intervals

    def hyperoptimize(
        self,
        n_trials: int = 64,
        rolling_validation: bool = True,
        train_subset: Union[None, pd.core.series.Series] = None,
    ) -> dict:
        """Search for hyperparameters with greater out-of-sample performance."""


class SurvivalModeler(Modeler):
    """Forecast probabilities of being observed in future periods.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast.
        allow_gaps (bool): Whether or not observations should be included for
            training and evaluation if there is a period without an observation
            between the period of observations and the last period of the
            given time horizon.
    """

    def __init__(self, **kwargs):
        """Initialize the SurvivalModeler.

        Args:
            **kwargs: Arguments to Modeler.__init__().
        """
        super().__init__(**kwargs)
        self.objective = "binary"
        self.num_class = 1

    def evaluate(
        self,
        subset: Union[None, pd.core.series.Series] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:
        """Tabulate model performance metrics.

        Args:
            subset: A Boolean Series that is True for observations over which
                the metrics will be computed. If None, default to all test
                observations in the earliest period of the test set.
            threshold_positive: None, "predicted", or a value in [0, 1]
                representing the minimum predicted probability considered to be a
                positive prediction; Specify "predicted" to use the predicted share
                positive in each time horizon. Overridden by share_positive.
            share_positive: None, "predicted", or a value in [0, 1] representing
                the share of observations with the highest predicted probabilities
                considered to have positive predictions. Specify "predicted" to use
                the predicted share positive in each time horizon. Probability ties
                may cause share of positives in output to exceed given value.
                Overrides threshold_positive.

        Returns:
            A DataFrame containing, for the binary outcomes of survival to each
            lead length, area under the receiver operating characteristic
            curve (AUROC), predicted and actual shares of observations with an
            outcome of True, and all elements of the confusion matrix. Also
            includes concordance index over the restricted mean survival time.
        """
        if subset is None:
            subset = self.data[self.test_col] & (
                self.data[self.period_col]
                == self.data[self.data[self.test_col]][self.period_col].min()
            )
        predictions = self.predict(subset=subset, cumulative=(not self.allow_gaps))
        metrics = []
        lead_lengths = np.arange(self.n_intervals) + 1
        for lead_length in lead_lengths:
            actuals = self.label_data(lead_length - 1)[subset].reset_index()
            actuals = actuals[actuals[self.max_lead_col] >= lead_length]
            weights = actuals[self.weight_col].values if self.weight_col else None
            actuals = actuals["_label"]
            metrics.append(
                compute_metrics_for_binary_outcome(
                    actuals,
                    predictions[:, lead_length - 1][actuals.index],
                    threshold_positive=threshold_positive,
                    share_positive=share_positive,
                    weights=weights,
                )
            )
        metrics = pd.DataFrame(metrics, index=lead_lengths)
        metrics.index.name = "Lead Length"
        metrics["Other Metrics:"] = ""
        if (not self.allow_gaps) and (self.weight_col is None):
            concordance_index_value = concordance_index(
                self.data[subset][[self.duration_col, self.max_lead_col]].min(axis=1),
                np.sum(predictions, axis=-1),
                self.data[subset][self.event_col],
            )
            metrics["C-Index"] = np.where(
                metrics.index == 1, concordance_index_value, ""
            )
        metrics = metrics.dropna()
        return metrics

    def forecast(self) -> pd.core.frame.DataFrame:
        """Tabulate survival probabilities for most recent observations."""
        columns = [
            str(i + 1) + "-period Survival Probability" for i in range(self.n_intervals)
        ]
        return pd.DataFrame(
            self.predict(
                subset=self.data[self.predict_col], cumulative=(not self.allow_gaps)
            ),
            columns=columns,
            index=(
                self.data[self.config["INDIVIDUAL_IDENTIFIER"]][
                    self.data[self.predict_col]
                ]
            ),
        )

    def tabulate_survival_by_quantile(
        self, n_quantiles: int, subset: Union[None, pd.core.series.Series] = None
    ) -> pd.core.frame.DataFrame:
        """Count number survived, grouped by survival probability.

        Args:
            subset: A Boolean Series that is True for observations over which
                the counts will be computed. If None, default to all
                observations.
            n_quantiles: Number of groups. Groups will contain the same number
                of observations except to keep observations with the same
                predicted probability in the same group. Quantile 1 represents
                the individuals least likely to survive the given lead length.

        Returns:
            A DataFrame where each row contains the group size and number
            survived for each combination of lead length and group.
        """
        subset = default_subset_to_all(subset, self.data)
        predictions = self.predict(subset=subset, cumulative=(not self.allow_gaps))
        actual_durations = self.data[subset][
            [self.duration_col, self.max_lead_col]
        ].min(axis=1)
        lead_lengths = np.arange(self.n_intervals) + 1
        counts = []
        for lead_length in lead_lengths:
            actuals = (
                actual_durations[self.data[self.max_lead_col] >= lead_length]
                >= lead_length
            )
            quantiles = pd.qcut(
                predictions[:, lead_length - 1][
                    self.data[subset][self.max_lead_col] >= lead_length
                ],
                n_quantiles,
                labels=False,
                duplicates="drop",
            )
            actuals = actuals.groupby(quantiles).agg(["sum", "count"])
            actuals["Lead Length"] = lead_length
            counts.append(actuals)
        counts = pd.concat(counts)
        counts.index.name = "Prediction Quantile"
        counts.columns = ["Number Survived", "Observations", "Lead Length"]
        counts = counts.reset_index()
        counts = counts[
            ["Lead Length", "Prediction Quantile", "Number Survived", "Observations"]
        ]
        counts["Prediction Quantile"] = counts["Prediction Quantile"] + 1
        return counts

    def tabulate_retention_rates(
        self, lead_periods: int, time_ids: Union[None, pd.core.series.Series] = None
    ):
        """Produce a table of actual and predicted retention rates.

        Retention rate is the share of individuals observed the given number of
        periods prior who survived to the corresponding period.

        Args:
            lead_periods: The number of periods an individual must survive to
                be considered retained.
            time_ids: Values representing the period of each observation. If
                None, default to number of periods since first period in data.

        Returns:
            Observed retention rate, predicted retention rate for periods where
            actual retention is observed in training set, and predicted
            retention rate for periods where actual retention rate is not
            observed in training set.
        """
        if time_ids is None:
            time_ids = self.data[self.period_col]
        actual_retention_rate = (
            (self.data[self.duration_col] >= lead_periods).groupby(time_ids).mean()
        )
        fitted_retention_rate = (
            pd.Series(self.predict()[:, lead_periods - 1]).groupby(time_ids).mean()
        )
        retention_rates = pd.DataFrame([actual_retention_rate, fitted_retention_rate]).T
        retention_rates.columns = ["Actual Retention Rate", "Fitted Retention Rate"]
        retention_rates["Actual Retention Rate"].iloc[-lead_periods:] = np.nan
        period_length = retention_rates.index[1] - retention_rates.index[0]
        for _ in range(lead_periods):
            retention_rates.loc[retention_rates.index.max() + period_length] = np.nan
        retention_rates = retention_rates.shift(lead_periods)
        retention_rates["Forecast Retention Rate"] = np.nan
        untrained_periods = lead_periods + self.config.get("TEST_PERIODS", 0)
        retention_rates["Forecast Retention Rate"].iloc[
            -untrained_periods:
        ] = retention_rates["Fitted Retention Rate"].iloc[-untrained_periods:]
        retention_rates["Fitted Retention Rate"].iloc[-untrained_periods:] = np.nan
        retention_rates.loc[retention_rates.index.max() + period_length] = np.nan
        retention_rates.index.name = "Period"
        return retention_rates

    def subset_for_training_horizon(
        self, data: pd.DataFrame, time_horizon: int
    ) -> pd.DataFrame:
        """Return only observations where survival would be observed."""
        if self.allow_gaps:
            return data[data[self.max_lead_col] > time_horizon]
        return data[(data[self.duration_col] + data[self.event_col] > time_horizon)]

    def label_data(self, time_horizon: int) -> pd.DataFrame:
        """Return data with an indicator for survival for each observation."""
        data = self.data.copy()
        data[self.duration_col] = data[[self.duration_col, self.max_lead_col]].min(
            axis=1
        )
        if self.allow_gaps:
            ids = data[
                [self.config["INDIVIDUAL_IDENTIFIER"], self.config["TIME_IDENTIFIER"]]
            ]
            ids[self.config["TIME_IDENTIFIER"]] = (
                ids[self.config["TIME_IDENTIFIER"]] - time_horizon - 1
            )
            ids["_label"] = True
            data = data.merge(
                ids,
                how="left",
                on=[
                    self.config["INDIVIDUAL_IDENTIFIER"],
                    self.config["TIME_IDENTIFIER"],
                ],
            )
            data["_label"] = data["_label"].fillna(False)
        else:
            data["_label"] = data[self.duration_col] > time_horizon
        return data


class StateModeler(Modeler):
    """Forecast the future value of a feature conditional on survival.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast.
        state_col (str): The column representing the state to forecast.
        objective (str): The model objective appropriate for the outcome type;
            "multiclass" for categorical states and "regression" for numeric
            states.
        class_values (pd.Int64Index): The state categories.
        num_class (int): The number of state categories or, if the state is
            numeric, None.
    """

    def __init__(self, state_col, **kwargs):
        """Initialize the StateModeler.

        Args:
            state_col: The column representing the state to forecast.
            **kwargs: Arguments to Modeler.__init__().
        """
        super().__init__(**kwargs)
        self.state_col = state_col
        if self.data is not None:
            if self.state_col in self.categorical_features:
                self.objective = "multiclass"
                self.class_values = self.data[self.state_col].cat.categories
                self.num_class = len(self.class_values)
            elif self.state_col in self.numeric_features:
                self.objective = "regression"
                self.num_class = None
            else:
                raise ValueError("state_col not in features.")

    def evaluate(
        self, subset: Union[None, pd.core.series.Series] = None
    ) -> pd.core.frame.DataFrame:
        """Tabulate model performance metrics.

        Args:
            subset: A Boolean Series that is True for observations over which
                the metrics will be computed. If None, default to all test
                observations in the earliest period of the test set.

        Returns:
            A DataFrame containing, for each lead length, area under the
            receiver operating characteristic curve (AUROC) for categorical
            states, or, for numeric states, R-squared.
        """
        if subset is None:
            subset = self.data[self.test_col] & (
                self.data[self.period_col]
                == self.data[self.data[self.test_col]][self.period_col].min()
            )
        predictions = self.predict(subset=subset, cumulative=False)
        metrics = []
        lead_lengths = np.arange(self.n_intervals) + 1
        for lead_length in lead_lengths:
            actuals = self.subset_for_training_horizon(
                self.label_data(lead_length - 1)[subset].reset_index(), lead_length - 1
            )
            weights = actuals[self.weight_col].values if self.weight_col else None
            actuals = actuals["_label"]
            if self.objective == "multiclass":
                actuals = pd.DataFrame(
                    {label: actuals == label for label in range(self.num_class)},
                    index=actuals.index,
                )
                metrics.append(
                    compute_metrics_for_categorical_outcome(
                        actuals,
                        predictions[:, :, lead_length - 1].T[actuals.index],
                        weights=weights,
                    )
                )
            else:
                metrics.append(
                    compute_metrics_for_numeric_outcome(
                        actuals,
                        predictions[:, lead_length - 1][actuals.index],
                        weights=weights,
                    )
                )
        metrics = pd.DataFrame(metrics, index=lead_lengths)
        metrics.index.name = "Lead Length"
        return metrics

    def forecast(self) -> pd.core.frame.DataFrame:
        """Tabulate state probabilities for most recent observations."""
        forecasts = self.predict(subset=self.data[self.predict_col], cumulative=False)
        if self.objective == "multiclass":
            columns = [
                str(i + 1) + "-period State Probabilities"
                for i in range(self.n_intervals)
            ]
            ids = np.repeat(
                self.data[self.config["INDIVIDUAL_IDENTIFIER"]][
                    self.data[self.predict_col]
                ],
                forecasts.shape[0],
            )
            states = np.tile(self.class_values, forecasts.shape[1])
            forecasts = np.reshape(
                forecasts,
                (forecasts.shape[0] * forecasts.shape[1], forecasts.shape[2]),
                order="F",
            )
            index = pd.MultiIndex.from_arrays(
                [ids, states],
                names=[
                    self.config["INDIVIDUAL_IDENTIFIER"],
                    f"Future {self.state_col}",
                ],
            )
            forecasts = pd.DataFrame(forecasts, columns=columns, index=index)
        else:
            columns = [
                str(i + 1) + f"-period {self.state_col} Forecast"
                for i in range(self.n_intervals)
            ]
            index = self.data[self.config["INDIVIDUAL_IDENTIFIER"]][
                self.data[self.predict_col]
            ]
            forecasts = pd.DataFrame(forecasts, columns=columns, index=index)
        return forecasts

    def subset_for_training_horizon(
        self, data: pd.DataFrame, time_horizon: int
    ) -> pd.DataFrame:
        """Return only observations where the future state is observed."""
        if self.allow_gaps:
            return data[data[self.max_lead_col] > time_horizon].dropna(
                subset=["_label"]
            )
        return data[(data[self.duration_col] > time_horizon)]

    def label_data(self, time_horizon: int) -> pd.Series:
        """Return data with the future state for each observation."""
        data = self.data.copy()
        data[self.duration_col] = data[[self.duration_col, self.max_lead_col]].min(
            axis=1
        )
        ids = data[
            [
                self.config["INDIVIDUAL_IDENTIFIER"],
                self.period_col,
                self.state_col,
            ]
        ]
        ids[self.period_col] = ids[self.period_col] - time_horizon - 1
        ids = ids.rename({self.state_col: "_label"}, axis=1)
        if self.objective == "multiclass":
            ids["_label"] = ids["_label"].cat.codes
        data = data.merge(
            ids,
            how="left",
            on=[self.config["INDIVIDUAL_IDENTIFIER"], self.period_col],
        )
        return data


class ExitModeler(StateModeler):
    """Forecast the circumstance of exit conditional on exit."""

    def __init__(self, exit_col, **kwargs):
        """Initialize the ExitModeler.

        Args:
            exit_col: The column representing the exit circumstance to forecast.
            **kwargs: Arguments to Modeler.__init__().
        """
        super().__init__(exit_col, **kwargs)
        if self.objective == "multiclass":
            self.class_values = self.data[
                (self.data["_duration"] == 0) & (self.data["_event_observed"] == True)
            ][self.state_col].unique()
            self.data[self.state_col] = self.data[
                self.state_col
            ].cat.reorder_categories(
                list(self.class_values)
                + [
                    cat
                    for cat in self.data[self.state_col].cat.categories
                    if cat not in self.class_values
                ]
            )
            self.num_class = len(self.class_values)
        self.exit_col = self.state_col
        if self.data is not None:
            if self.state_col in self.categorical_features:
                self.categorical_features.remove(self.state_col)
            elif self.state_col in self.numeric_features:
                self.numeric_features.remove(self.state_col)

    def subset_for_training_horizon(
        self, data: pd.DataFrame, time_horizon: int
    ) -> pd.DataFrame:
        """Return only observations where exit is observed at the given time horizon."""
        if self.allow_gaps:
            ids = data[
                [self.config["INDIVIDUAL_IDENTIFIER"], self.config["TIME_IDENTIFIER"]]
            ]
            ids[self.config["TIME_IDENTIFIER"]] = (
                ids[self.config["TIME_IDENTIFIER"]] - time_horizon - 1
            )
            ids["exit"] = False
            data = data.merge(
                ids,
                how="left",
                on=[
                    self.config["INDIVIDUAL_IDENTIFIER"],
                    self.config["TIME_IDENTIFIER"],
                ],
            )
            data = data[data["exit"].isna() & (data[self.max_lead_col] > time_horizon)]
            data = data.drop("exit", axis=1)
        else:
            data = data[
                data[self.event_col] & (data[self.duration_col] == time_horizon)
            ]
        return data

    def label_data(self, time_horizon: int) -> pd.Series:
        """Return data with the exit circumstance for each observation."""
        data = self.data.copy()
        data[self.duration_col] = data[[self.duration_col, self.max_lead_col]].min(
            axis=1
        )
        if self.objective == "multiclass":
            temp_exit_col = data[self.exit_col]
            data[self.exit_col] = data[self.exit_col].cat.codes
        data["_label"] = data.groupby(
            [self.config["INDIVIDUAL_IDENTIFIER"], self.spell_col]
        )[self.exit_col].transform("last")
        if self.objective == "multiclass":
            data[self.exit_col] = temp_exit_col
        return data



"""Setup for FIFE: Finite-Interval Forecasting Engine."""

from setuptools import setup

with open("README.md", "r") as f:
    README = f.read()

setup(
    name="fife",
    version="1.5.1",
    description=(
        "Finite-Interval Forecasting Engine: Machine learning models "
        "for discrete-time survival analysis and multivariate time series "
        "forecasting"
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/IDA-HumanCapital/fife",
    project_urls={
        "Bug Tracker": "https://github.com/IDA-HumanCapital/fife/issues",
        "Source Code": "https://github.com/IDA-HumanCapital/fife",
        "Documentation": "https://fife.readthedocs.io/en/latest",
    },
    author="Institute for Defense Analyses",
    author_email="humancapital@ida.org",
    license="AGPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: "
        "GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    packages=["fife"],
    install_requires=[
        "dask[delayed]",
        "ipython",
        "keras",
        "lifelines",
        "lightgbm",
        "matplotlib",
        "numpy",
        "optuna",
        "pandas",
        "seaborn",
        "scikit-learn",
        "tensorflow",
    ],
    extras_require={"shap": ["shap"]},
    entry_points={
        "console_scripts": [
            "fife=fife.__main__:main",
        ]
    },
)



"""FIFE modeler based on Pandas, which tabulates interacted fixed effects."""

from typing import Union
from warnings import warn

from fife.base_modelers import (
    default_subset_to_all,
    Modeler,
    SurvivalModeler,
    StateModeler,
    ExitModeler,
)
import numpy as np
import pandas as pd


class IFEModeler(Modeler):
    """Predict with mean of training observations with same values.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of one-period intervals any
            individual is observed to survive.
        model (pd.core.frame.DataFrame): Survival rates for each combination of
            categorical values in the training data.
    """

    def train(self) -> pd.core.frame.DataFrame:
        """Compute the mean of the outcome for each combination of categorical values."""
        cell_rates = pd.DataFrame()
        train_subset = (
            ~self.data[self.validation_col]
            & ~self.data[self.test_col]
            & ~self.data[self.predict_col]
        )
        for time_horizon in range(self.n_intervals):
            data = self.label_data(time_horizon)
            data = data[train_subset]
            data = self.subset_for_training_horizon(data, time_horizon)
            grouping = self.categorical_features
            if self.objective == "multiclass":
                if self.weight_col:
                    grouped_sums = data.groupby(self.categorical_features + ["_label"])[
                        self.weight_col
                    ].sum()
                    cell_rates[
                        time_horizon + 1
                    ] = grouped_sums / grouped_sums.reset_index().groupby(
                        self.categorical_features
                    )[
                        self.weight_col
                    ].transform(
                        "sum"
                    )
                else:
                    grouped_sums = data.groupby(self.categorical_features + ["_label"])[
                        "_label"
                    ].count()
                    cell_rates[time_horizon + 1] = (
                        grouped_sums
                        / data.groupby(self.categorical_features)["_label"].count()
                    )
            else:
                if self.weight_col:
                    data["_label"] = data["_label"] * data[self.weight_col]
                    grouped_sums = data.groupby(self.categorical_features)[
                        "_label", self.weight_col
                    ].sum()
                    cell_rates[time_horizon + 1] = (
                        grouped_sums["_label"] / grouped_sums[self.weight_col]
                    )
                else:
                    cell_rates[time_horizon + 1] = data.groupby(
                        self.categorical_features
                    )["_label"].mean()
        return cell_rates

    def predict(
        self, subset: Union[None, pd.core.series.Series] = None, cumulative: bool = True
    ) -> np.ndarray:
        """Map observations to outcome means from their categorical values.

        Map observations with a combination of categorical values not seen in
        the training data to the mean of the outcome in the training set.

        Args:
            subset: A Boolean Series that is True for observations for which
                predictions will be produced. If None, default to all
                observations.
            cumulative: If True, will produce cumulative survival
                probabilities. If False, will produce marginal survival
                probabilities (i.e., one minus the hazard rate).

        Returns:
            A numpy array of outcome means by observation and lead
            length.
        """
        subset = default_subset_to_all(subset, self.data)
        predictions = self.data[subset].merge(
            self.model, how="left", on=self.categorical_features
        )
        predictions = predictions[self.model.columns]
        predictions = predictions.to_numpy()
        if self.objective == "multiclass":
            predictions = predictions.reshape(
                (self.num_class, subset.sum(), self.n_intervals), order="F"
            )
        if cumulative:
            predictions = np.cumprod(predictions, axis=-1)
        return predictions

    def save_model(self, file_name: str = "IFE_Model", path: str = "") -> None:
        """Save the pandas DataFrame model to disk."""
        self.model.to_csv(path + file_name + ".csv")

    def hyperoptimize(self, **kwargs) -> dict:
        """Returns None for InteractedFixedEffectsModeler, which does not have hyperparameters"""
        warn(
            "Warning: InteractedFixedEffectsModeler does not have hyperparameters to optimize."
        )
        return None


class IFESurvivalModeler(IFEModeler, SurvivalModeler):
    """Predict with survival rate of training observations with same values."""

    pass


class InteractedFixedEffectsModeler(IFESurvivalModeler):
    """Deprecated alias for IFESurvivalModeler"""

    warn(
        "The name 'InteractedFixedEffectsModeler' is deprecated. "
        "Please use 'IFESurvivalModeler' instead.",
        DeprecationWarning,
    )


class IFEStateModeler(IFEModeler, StateModeler):
    """Forecast the future value of a feature conditional on survival using the mean of observations with the same values."""

    pass


class IFEExitModeler(IFEModeler, ExitModeler):
    """Forecast the circumstance of exit conditional on exit using the mean of observations with the same values."""

    pass
